import re
import gremlin_python
from functools import wraps
from datetime import datetime

str_time = lambda: datetime.now().strftime('%Y-%m-%d at %H:%M:%S.%f')


class classinstancemethod:
    def __init__(self, method, instance=None, owner=None):
        self.method = method
        self.instance = instance
        self.owner = owner

    def __get__(self, instance, owner=None):
        return wraps(self.method)(type(self)(self.method, instance=instance, owner=owner))

    def __call__(self, *args, **kwargs):
        return self.method(self.instance, self.owner, *args, **kwargs)


def camel_to_snake(string, sep='_', lower=True):
    string = re.sub('(.)([A-Z][a-z]+)', fr'\1{sep}\2', string)
    string = re.sub('(.)([0-9]+)', fr'\1{sep}\2', string)
    string = re.sub('([a-z0-9])([A-Z])', fr'\1{sep}\2', string)
    if lower:
        return string.lower()
    else:
        return string


def check_structure(structure, kwproperties):
    missing_args = []
    for var in structure:
        if var not in kwproperties:
            missing_args.append(var)

    if missing_args:
        m = f"missing {len(missing_args)} required positional argument{'' if len(missing_args) == 1 else 's'}: "
        raise TypeError(m + ', '.join(missing_args))


def get_traversal(x) -> 'Traversal':
    from janusgraphy.traversal import Traversal  # late import to avoid circular import
    from janusgraphy.query import Query

    if type(x) is gremlin_python.structure.graph.Vertex:
        x = Traversal(t_query=[['g'], ['V', [x.id]]])
    elif type(x) is gremlin_python.structure.graph.Edge:
        edge_id = x.id['@value']['value']
        x = Traversal(t_query=[['g'], ['E', [edge_id]]])
    #elif type(x) is list:
    #    x = Traversal([['g'], [mapping[type(x)], [] + list(map(lambda gobj: gobj.id, x))]])
    elif type(x) is Traversal:
        x = x.clone()
    elif type(x) is Query:
        x = Traversal(t_query=x.query)
    else:
        print('Unrecognized type in get_traversal:', type(x))
    return x



'''
Please ignore this class
'''

traversal_verbose = False


def pformatter(l):
    return map(lambda i: f'==>{str(i)}', l)


def pprinter(l):
    print('\n'.join(pformatter(l)))


#connect()

qrm = lambda q: map(GraphObject.instantiate, client.submit(q).all().result())
qr = lambda q: list(qrm(q))
qrl = qr
qrp = lambda q: pprinter(qrm(q))


class Cls:
    def __repr__(*a, **k):
        import os
        os.system('clear')
        return ""


clear = Cls()

# returns a dict of {name:devicename, status:laststatus} from the selected client
'''
g.V().has("name", "Neuropsicocentro") # gets client
.in("From").has("type", "smartplug").order().by("name", incr) # gets devices sorted by name
.project("name", "status") # prepare projection: a list of dicts of type {name:value1, status:value2}
.by(values("name")) # sets first value
.by(                # sets second value
    coalesce(       # returns the first successfull query
        __.in("From").order().by("date", decr).limit(1).values("message"), # try to return this
         constant("not found")))'   # returns this if the first failed

or

q = Query(verbose=True).filter_by_property(name='Neuropsicocentro')
q.through_incoming_edge("From").filter_by_property(type="smartplug").sort('name')
status_q = Query.relation().through_incoming_edge('From').sort('date', descending=True)[0].get_values('message')
q.to_dictionary(name=Query.relation().get_values('name'), status=Query.relation().get(status_q, '-'))

pprinter(q.fetch_all())
'''

# returns a dict of {subtype:logCount}
"g.V().hasLabel('Log').out('From').has('type', 'smartplug').values('subtype').groupCount()"
#q = Query().filter_by_property(Label="Log").through_outgoing_edge('From').filter_by_property(
#    type='smartplug').get_values('name').group_and_count()
#pprinter(q.fetch_all())


__version__ = '0.1.0'

import re
import gremlin_python
from gremlin_python.driver.client import Client

from janusgraphy.traversal import Traversal
from janusgraphy.graph_object import GraphObject
from janusgraphy import helpers

traversal_verbose = False
master_client = None


class Edge(GraphObject):
    pass


class Vertex(GraphObject):
    def add_to_graph(self):
        if not self.in_graph():
            self.graph_value = add_vertex(self.Label, **self.Properties).run()[0]
        return self.graph_value

    def add_edge(self, edge, to, two_way=False, **kwproperties):
        self.add_to_graph()
        to.add_to_graph()

        if hasattr(edge, 'Label'):
            edge = edge.Label

        if hasattr(edge, 'Structure'):
            helpers.check_structure(edge.Structure, kwproperties)

        traversal = add_edge(self.graph_value, edge, to.graph_value, both_directions=two_way, **kwproperties)
        return run(traversal)


class LoggedVertex(Vertex):
    link_start_property_name = 'link_start'
    link_end_property_name = 'link_end'

    def remove_from_graph(self, **kwargs):
        edges = self.query().to_edge()
        if self.link_end_property_name in kwargs:
            add_property(edges, **{self.link_end_property_name: kwargs[self.link_end_property_name]}).run()
        else:
            add_property(edges, **{self.link_end_property_name: helpers.str_time()}).run()

    def add_edge(self, *args, **kwargs):
        if self.link_start_property_name in kwargs:
            super().add_edge(*args, **kwargs)
        else:
            super().add_edge(*args, **{self.link_start_property_name: helpers.str_time()}, **kwargs)


def run(traversal, verbose=None) -> list:
    if type(traversal) is Traversal:
        traversal = [traversal]

    r = []
    for trav in traversal:
        r += trav.run(verbose=verbose)

    return r


def connect(url='ws://localhost:8182/gremlin', name='g') -> Client:
    return Client(url, name)


def connect_master(url='ws://localhost:8182/gremlin', name='g') -> Client:
    global master_client
    master_client = Client(url, name)
    return master_client


def add_property(target, **properties) -> Traversal:
    target = helpers.get_traversal(target)
    for key, value in properties.items():
        target.add('property', key, value)

    return target


def add_vertex(Label, **properties) -> Traversal:
    v = Traversal.graph().add('addV', Label)
    return add_property(v, **properties)


def add_edge(source, Label, destination, both_directions=False, **properties) -> list:
    def _add(_source, _label, _destination, _properties):
        _source = helpers.get_traversal(_source)
        _destination = helpers.get_traversal(_destination)

        _source.add('addE', _label).add('to', _destination)
        result = add_property(_source, **_properties)
        return result

    r = [_add(_source=source, _label=Label, _destination=destination, _properties=properties)]

    if both_directions:
        r += [_add(_source=destination, _label=Label, _destination=source, _properties=properties)]

    return r


from janusgraphy.query import Query


def get_vertex_from_id(id, Label=None):
    return Query.from_vertex_id(id, Label).fetch_first()


def get_edge_from_id(id, Label=None):
    return Query.from_edge_id(id, Label).fetch_first()


import gremlin_python
import dataclasses


@dataclasses.dataclass
class TraversalLiteral:
    value: str


class Traversal:
    def __init__(self, t_query=None):
        #print('base query:', t_query)
        if type(t_query) in [str, int, float]:
            t_query = Traversal().add('constant', t_query).query
        self.query = (t_query or [])

    def add(self, *args):
        '''
        t.add('g') = g
        t.add('g', None) = g()
        t.add('g', []) = g()
        t.add('g', 'a') = g('a')
        t.add('g', 'a', 'b') = g('a', 'b')
        '''
        to_add = [args[0]]
        if len(args) > 1:
            if len(args) == 2 and args[1] is None:
                to_add.append([])
            else:
                to_add.append(args[1:])

        self.query.append(to_add)
        return self

    @staticmethod
    def graph():
        t = Traversal()
        t.add('g')
        return t

    @staticmethod
    def vertices():
        t = Traversal.graph()
        t.add('V', None)
        return t

    def __add__(self, other):
        r = None
        if type(other) is type(self):
            r = self.query + other.query

        elif type(other) is list:
            r = self.query + other

        elif type(other) is str:
            r = self.query + [other]

        return Traversal(r)

    def __radd__(self, other):
        r = None
        if type(other) is type(self):
            r = other.traversal_query + self.query

        elif type(other) is list:
            r = other + self.query

        elif type(other) is str:
            r = [other] + self.query

        return Traversal(r)

    def get_traversal_string(self) -> str:
        def dot_splt(q) -> str:
            rd = ''
            for part in q:
                rd += '.' + part[0]
                if len(part) > 1:
                    if len(part) == 2 and part[1] == []:
                        rd += '()'
                    else:
                        rd += '(' + comma_split(part[1]) + ')'

            return rd[1:]

        def comma_split(q) -> str:
            rc = []
            for part in q:
                if type(part) is TraversalLiteral:
                    rc.append(part.value)

                elif type(part) is str:
                    rc.append(f'"{part}"')

                elif type(part) is bool:
                    rc.append('true' if part else 'false')

                elif type(part) in [int, gremlin_python.statics.long]:
                    rc.append(str(part))

                elif type(part) is float:
                    rc.append(f'{str(part)}f')

                elif type(part) is type(self):
                    rc.append(part.get_traversal_string())

                elif type(part) is list:
                    rc.append(dot_splt(part))

                else:
                    print('unrecognized type:', str(type(part)))

            return ', '.join(rc)

        r = dot_splt(self.query)
        return r

    def run(self, client=None, verbose=False) -> list:
        ts = self.get_traversal_string().replace('\n', '\\n')

        if client is None:
            from janusgraphy import master_client
            client = master_client

        if verbose:
            print('traversal -->', ts)

        r1 = client.submit(ts)
        r2 = r1.all()
        r3 = r2.result()
        return r3

    def clone(self):
        from copy import deepcopy
        return deepcopy(self)

    @staticmethod
    def edges():
        return Traversal.graph().add('E', None)

    def __repr__(self):
        return f'<Traversal: {self.get_traversal_string()}>'



from janusgraphy.traversal import Traversal, TraversalLiteral
from janusgraphy.graph_object import GraphObject
from janusgraphy import traversal_verbose, run


class Query():
    def __init__(self, query_space=None, verbose=None, return_clones=None):
        if query_space is None:
            query_space = Traversal.vertices()
        self.query = query_space
        self.verbose = verbose

    def __getitem__(self, s):
        start, stop = 0, 0
        if isinstance(s, slice):
            start = s.start
            stop = s.stop
        else:
            start = s
            stop = s + 1
            # could add step 'limit'
            # self.query.add("limit", s)

        self.query.add("range", start, stop)
        return self

    def __repr__(self):
        return f'<Query: {self.query.get_traversal_string()}>'

    @staticmethod
    def from_vertex_id(vertex_id: int, Label=None):
        query = Query(query_space=Traversal(t_query=[['g'], ['V', [vertex_id]]]))
        if Label:
            query.filter_by_property(Label=Label)
        return query

    @staticmethod
    def from_edge_id(edge_id: int, Label=None):
        query = Query(query_space=Traversal(t_query=[['g'], ['E', [edge_id]]]))
        if Label:
            query.filter_by_property(Label=Label)
        return query

    @staticmethod
    def relation(return_clones=None):
        return Query(query_space=Traversal().add('__'), return_clones=return_clones)

    # comparisons
    # Usage: filter_by_property(age=greater_or_equal(18), name=regex('[a-zA-Z]*'))
    @staticmethod
    def not_equal(val, return_clones=None):
        return Traversal().add('neq', val)

    @staticmethod
    def contains_word(val):
        return Traversal().add('textContains', val)

    @staticmethod
    def contains_regex_word(val):
        if type(val) is str:
            val = val.replace('*', '.*')
        return Traversal().add('textContainsRegex', val)

    @staticmethod
    def contains_similar_word(val):
        return Traversal().add('textContainsFuzzy', val)

    @staticmethod
    def regex(val):
        if type(val) is str:
            val = val.replace('*', '.*')
        return Traversal().add('textRegex', val)

    @staticmethod
    def is_similar(val):
        return Traversal().add('textFuzzy', val)

    @staticmethod
    def greater(val):
        return Traversal().add('gt', val)

    @staticmethod
    def greater_or_equal(val):
        return Traversal().add('gte', val)

    @staticmethod
    def less(val):
        return Traversal().add('lt', val)

    @staticmethod
    def less_or_equal(val):
        return Traversal().add('lte', val)

    def set_property(self, **kwproperties):
        for key, value in kwproperties.items():
            self.query.add('property', key, value)
        return self

    def filter_by_property(self, *properties, id=None, Label=None, **kwproperties):

        if Label:
            if type(Label) is type(self):  # Why have I done this?? When does this happen?
                Label = Label.query

            elif hasattr(Label, 'Label'):  # to handle passing the GraphObject type
                Label = Label.Label

            self.query.add('hasLabel', Label)

        if id:
            if type(id) is type(self):  # why have I done this?? When does this happen?
                id = id.query

            elif hasattr(id, 'graph_value'):  # to handle passing the GraphObject type
                id = id.graph_value.id

            self.query.add('has', TraversalLiteral('id'), id)

        for key, value in kwproperties.items():
            self.query.add('has', key, value)

        for prop in properties:
            self.query.add('has', prop)

        return self

    def filter_by_relation(self, relation):
        import uuid
        temp_tag = str(uuid.uuid4())
        self.tag_as(temp_tag)
        if relation.query.query[0] == ['__']:
            relation.query.query = relation.query.query[1:]
        self.query += relation.query
        self.select_with_tag(temp_tag)
        return self.remove_duplicates()

    def remove_duplicates(self):
        self.query.add('dedup', None)
        return self

    def filter_by_edge(self, *properties):
        for any_edge in properties:
            if hasattr(any_edge, 'Label'):
                any_edge = any_edge.Label
            self.query.add('bothE', any_edge).add('inV', None)

        return self

    def filter_by_incoming_edge(self, *properties):
        for in_edge in properties:
            if hasattr(in_edge, 'Label'):
                in_edge = in_edge.Label
            self.query.add('inE', in_edge).add('inV', None)

        return self

    def filter_by_outgoing_edge(self, *properties):
        for out_edge in properties:
            if hasattr(out_edge, 'Label'):
                out_edge = out_edge.Label
            self.query.add('inE', out_edge).add('inV')

        return self

    def tag_as(self, tag):
        self.query.add('as', tag)
        return self

    def select_with_tag(self, tag):
        self.query.add('select', tag)
        return self

    def through_outgoing_edge(self, Label=None):
        if Label:
            if hasattr(Label, 'Label'):
                Label = Label.Label
            self.query.add('out', Label)
        else:
            self.query.add('out', None)
        return self

    def through_incoming_edge(self, Label=None):
        if Label:
            if hasattr(Label, 'Label'):
                Label = Label.Label
            self.query.add('in', Label)
        else:
            self.query.add('in', None)
        return self

    def through_edge(self, Label=None):
        if Label:
            if hasattr(Label, 'Label'):
                Label = Label.Label
            self.query.add('both', Label)
        else:
            self.query.add('both', None)
        return self

    def to_outgoing_edge(self, Label=None):
        if Label:
            if hasattr(Label, 'Label'):
                Label = Label.Label
            self.query.add('outE', Label)
        else:
            self.query.add('outE', None)
        return self

    def to_incoming_edge(self, Label=None):
        if Label:
            if hasattr(Label, 'Label'):
                Label = Label.Label
            self.query.add('inE', Label)
        else:
            self.query.add('inE', None)
        return self

    def to_edge(self, Label=None):
        if Label:
            if hasattr(Label, 'Label'):
                Label = Label.Label
            self.query.add('bothE', Label)
        else:
            self.query.add('bothE', None)
        return self

    def to_destination_vertex(self):
        self.query.add('inV', None)
        return self

    def to_source_vertex(self):
        self.query.add('outV', None)
        return self

    def sort(self, *args, descending=False):
        order = 'decr' if descending else 'incr'
        if not args and not descending:
            self.query.add('order', None)
        else:
            self.query.add('order', None).add('by', *args, [[order]])

        return self

    def get(self, *queries):
        """Gets the first successful query
        """
        queries = list(queries)
        for i, q in enumerate(queries):
            if type(q) is type(self):
                queries[i] = q.query
            elif type(q) is not Traversal:
                queries[i] = Traversal(q)

        self.query.add('coalesce', *queries)
        return self

    def get_values(self, *values):
        self.query.add('values', *values)
        return self

    def group_and_count(self, key=None):
        '''
        counts the number of occourences of each item and 
        groups them into a dictionary.
        Same as group_by(key, Query.relation().count())
        '''
        self.query.add("groupCount", key)
        return self

    def group_by(self, key, value=None):
        """Groups the traversals and returns the
        value specified by 'value' parameter

        The traversals are first grouped and then the
        value is applied over each of the groups
        """
        if type(key) is type(self):
            key = key.query

        self.query.add('group', None).add('by', key)
        if value is not None:
            if type(value) is type(self):
                value = value.query

            self.query.add('by', value)

        return self

    def count(self):
        self.query.add('count', None)
        return self

    def delete(self):
        self.query.add('drop', None)
        return self

    def to_dictionary(self, **kwkeys):
        keys = list(kwkeys)
        self.query.add('project', *keys)
        for k in keys:
            trav = kwkeys[k]
            if type(trav) is type(self):
                trav = trav.query
            self.query.add('by', trav)
        return self

    @staticmethod
    def _get_properties(obj: GraphObject, props: list):
        ret_list = []
        for prop in props:
            r = obj.Properties.get(prop, None)
            if r is not None:
                ret_list.append(r)
        return ret_list

    def fetch_all(self, *args):
        try:
            if not args:
                r = self.fetch_raw()
                return map(GraphObject.instantiate, r)
            else:
                return self.get_values(*args).fetch_raw()
        except Exception as e:
            print(e)
            raise

    def fetch_raw(self, verbose_override=None):
        if verbose_override is not None:
            verb = verbose_override
        else:
            verb = self.verbose

        return run(self.query, verbose=verb)

    def fetch_first(self, *args, ensure_one=False):
        try:
            r = self.fetch_all(*args)
            #if len(r) > 0
            return r.__next__()

        except Exception as e:
            print(e)
            raise

    def clone(self):
        nq = Query(query_space=self.query.clone(), verbose=self.verbose)
        return nq

    #def fetch_one(self):
    #    return GraphObject.instantiate(self.query.next())



import gremlin_python
from janusgraphy import helpers

traversal_verbose = False


class GraphObjectMeta(type):
    known_objects = dict()

    def __new__(cls, clsname, superclasses, attributedict):
        if traversal_verbose:
            print("clsname: ", clsname)
            print("superclasses: ", superclasses)
            print("attributedict: ", attributedict)
        new_class = type.__new__(cls, clsname, superclasses, attributedict)
        return new_class

    def __init__(cls, clsname, superclasses, attributedict):
        if traversal_verbose:
            print("clsname: ", clsname)
            print("superclasses: ", superclasses)
            print("attributedict: ", attributedict)

        if not attributedict.get('Label'):
            cls.Label = helpers.camel_to_snake(clsname, sep=' ', lower=False)

        if cls.Label in GraphObjectMeta.known_objects:
            raise Exception(f"Class label '{cls.Label}' is already defined")

        GraphObjectMeta.known_objects[cls.Label] = cls

    '''def __new__(cls, clsname, superclasses, attributedict):
        print("clsname: ", clsname)
        print("superclasses: ", superclasses)
        print("attributedict: ", attributedict)
        new_class = type.__new__(cls, clsname, superclasses, attributedict)
        new_class.Label = clsname.lower()
        return new_class
    '''


class GraphObject(metaclass=GraphObjectMeta):
    """The base graph object

    The object's label is converted by default from CamelCase to snake_case.
    When extending, the object must implement add_to_graph method and can set
    a custom label through Label property and some mandatory properties

    The kwarguments passed in the __init__ method will be set in the database
    as properties
    
    Raises:
        NotImplementedError: on add_to_graph
    """
    #known_objects = dict()
    Label = None  # set by metaclass but can be set by extending class
    Structure = None  # properties that are mandatory. Can be set by extending class

    Properties: dict  # current object properties
    graph_value: 'Any' = None  # the value in the graph

    def __new__(cls, *args, add_to_graph: bool = True, set_properties: bool = False, **kwproperties):
        self = super(GraphObject, cls).__new__(cls)
        self.graph_value = None

        if set_properties:
            self.__set_properties(**kwproperties)
            if add_to_graph:
                self.add_to_graph()

        return self

    def __init__(self, *args, add_to_graph: bool = True, **kwproperties):
        """Initializes a GraphObject

        The kwarguments passed are set as properties of the database object
        (the logic that sets the properties is in __new__)
        
        Keyword Arguments:
            add_to_graph {bool} -- Whether to add the object to the graph already (default: {True})
        """
        self.__set_properties(**kwproperties)

        if add_to_graph:
            self.add_to_graph()

    def __set_properties(self, **properties):
        if self.Structure:
            helpers.check_structure(self.Structure, properties)

        self.Properties = properties

    @staticmethod
    def instantiate(obj):
        gremlin_vertex = gremlin_python.structure.graph.Vertex
        gremlin_edge = gremlin_python.structure.graph.Edge

        if type(obj) in [gremlin_vertex, gremlin_edge]:
            traversal = helpers.get_traversal(obj)
            label = obj.label

            values = (traversal + [['valueMap', []]]).run()[0]
            if type(obj) is gremlin_vertex:
                values = {k: v[0] for k, v in values.items()}

            cls = GraphObject.known_objects.get(label, GraphObject)
            result = cls.__new__(cls, add_to_graph=False, set_properties=True, **values)
            result.graph_value = obj
            return result

        if type(obj) is list:
            for i, e in enumerate(obj):
                obj[i] = GraphObject.instantiate(e)

            return obj

        if type(obj) is dict:
            inst = GraphObject.instantiate
            result = dict()
            for k, v in obj.items():
                result[inst(k)] = inst(v)

            return result

        else:
            return obj

    @helpers.classinstancemethod
    def query(self=None, cls=None, verbose=None) -> 'Query':
        from janusgraphy.query import Query  # late imports to avoid circular imports
        from janusgraphy.traversal import Traversal

        if self:
            return Query(query_space=helpers.get_traversal(self.graph_value), verbose=verbose)
        else:
            return Query(query_space=Traversal.vertices(), verbose=verbose).filter_by_property(Label=cls.Label)

    def remove_from_graph(self):
        if self.in_graph():
            helpers.get_traversal(self.graph_value).add('drop', None).run()
            return True
        else:
            return False

    def in_graph(self):
        return self.graph_value is not None

    def set_properties(self, **kwproperties):
        self.query().set_property(**kwproperties).fetch_raw()
        self.Properties.update(kwproperties)

    def __repr__(self):
        return f'<{self.Label}: {str(self.__dict__)}>'

    def add_to_graph(self):
        raise NotImplementedError()
